local M = {}

local tGroup = {} --Traffic group
--Both JBeam and config names (as long as they're in the right format) will work within these groups
tGroup.normal = {"autobello", "barstow", "burnside", "coupe", "etk800", "etkc", "etki", "fullsize", "hatch", "hopper", "legran", "midsize", "miramar", "moonhawk", "pessima", "pickup", "pigeon", "roamer", "sbr", "sunburst", "super", "van"}
tGroup.allBase = {}
tGroup.allAny = {}
tGroup.light = {"autobello/130_m.pc", "coupe", "hatch/DXi_M.pc", "midsize/LX_M.pc", "miramar", "pessima/HX_M.pc", "pigeon", "pigeon/boxed.pc"}
tGroup.heavy = {"citybus", "pickup/v8_4wd.pc", "pickup/beast.pc", "roamer/v8_4wd.pc", "roamer/offroad.pc", "semi", "semi/t75_dumptruck.pc", "van/medium.pc", "van/deliverytruck.pc"}
tGroup.america = {"barstow", "burnside", "coupe", "fullsize", "hatch", "legran", "midsize", "moonhawk", "pickup/ext_4wd.pc", "roamer", "sunburst", "van", "van/deliverytruck.pc"}
tGroup.europe = {"autobello", "autobello/130b_m.pc", "coupe", "etk800", "etkc", "etki", "etki/3000i_M_alt.pc", "hatch", "midsize", "pigeon", "super", "sbr"}
tGroup.asia = {"coupe", "hatch", "midsize", "miramar", "pessima", "pigeon", "sbr", "sunburst", "sunburst/custom.pc", "van/flatbed.pc"}
tGroup.americaPolice = {"coupe/police.pc", "fullsize/police.pc", "fullsize/bcpd.pc", "burnside/police.pc", "sunburst/police.pc", "sunburst/police_RS.pc", "roamer/sheriff.pc"}
tGroup.europePolice = {"super/polizia.pc", "etk800/police.pc", "etkc/police_highway_A.pc", "autobello/carabinieri.pc", "sunburst/gendarmerie.pc"}
tGroup.allPolice = {"coupe/police.pc", "super/polizia.pc", "etk800/police.pc", "fullsize/police.pc", "fullsize/bcpd.pc", "fullsize/unmarked.pc", "etkc/police_highway_A.pc", "autobello/carabinieri.pc", "burnside/police.pc", "sunburst/gendarmerie.pc", "sunburst/police.pc", "sunburst/police_RS.pc", "van/police.pc", "pickup/borderpatrol.pc", "roamer/sheriff.pc", "roamer/sheriff_offroad.pc"}
tGroup.automation = {}
tGroup.copyPlayer = {}
tGroup.custom = {"pigeon/600.pc"}
local paintColors = {{0.9, 0.05, 0.04, 1.35}, {0.5, 0, 0, 1.35}, {1, 0.3, 0, 1.35}, {0.95, 0.7, 0.1, 1.35}, {0.24, 0.48, 0.48, 1.35}, {0.0, 0.2, 0.7, 1.35}, {0.0, 0.05, 0.25, 1.35}, {1, 1, 1, 1.35}, {0.67, 0.67, 0.67, 1.35}, {0.5, 0.5, 0.5, 1.35}, {0.33, 0.33, 0.33, 1.35}, {0, 0, 0, 1.35}} --Standard traffic colors (future: use vehicle color table)

local player = {}
local traffic = {}
local Player = {}
local Traffic = {}

zSiren = 0 --Intentionally global (for now)
local queue = 0
local loopTimer = 0
local timeTick = 0.1 + 1e-6
local speedLimit = 14
local speedConvert = {3.6, 2.2369} --km/h, mph
local speedUnits = 1
local trafficTotal = 1
local spawnFactor = 1
local spawnPassFactor = 1.6
local spawnGapFactor = 1
local policeSpawn = 1.2
local damageUnit = 1000
local up = vec3(0, 0, 1)
local group = tGroup.normal
local groupName = "normal"
local trafficMode = "random"
local laneMode = "on"
local paintColorMode = "standard"
local metricSystem = true
local smartSpeed = false
local trafficReady = false
local policeReady = false
local policeAlert = false
local scatterCenter = true
local headlightsMode = true
local damageMode = true
local crawlSpeed = 3 --Slow vehicle speed
local arrestTime = 5 --Number of seconds until police arrest you
local evadeTime = 40 --Number of seconds away from police for evasion
local roadDrive = 0.2 --Minimum road drivability needed for spawn point
local nearDist = 40 --Near distance, used for traffic passing and interactions
local farDist = 400 --Far distance, used for traffic despawning
local pointDist = 200 --Start distance for point ahead of player vehicle; used for setting a spawn point
local readyDist = 100 --Proximity distance, used for a variety of traffic and police logic
local distMultiplier = 50 --Number used as a multiplier for despawn logic

function Player:new(obj, pos, vel, vec, dist, pair, weight, damage, state, timer)
	local self = setmetatable({}, Player)
	self.obj = obj or nil --The vehicle object itself
	self.pos = pos or nil --Vehicle position
	self.vel = vel or 0 --Vehicle velocity
	self.vec = vec or nil --Vehicle direction vector
	self.dist = dist or 0 --Vehicle distance to others
	self.pair = pair or {} --Nearest AI vehicle to player
	self.weight = weight or 0 --Vehicle weight
	self.damage = damage or 0 --Vehicle damage points
	self.state = state or nil --Action states; normal, stop
	self.timer = timer or {} --General timer for player
	return self
end

function Traffic:new(obj, pos, vel, vec, speed, dist, damage, state, role, action, lane, node, respawn, road, timer, drive)
	local self = setmetatable({}, Traffic)
	self.__index = self
	self.obj = obj or nil --The vehicle object itself
	self.pos = pos or nil --Vehicle position
	self.vel = vel or 0 --Vehicle velocity
	self.vec = vec or nil --Vehicle direction vector
	self.speed = speed or 0 --Vehicle speed to set as AI setSpeed
	self.dist = dist or {} --Vehicle distance to player
	self.damage = damage or {} --Vehicle damage points
	self.state = state or nil --Teleport states; none, normal, ready, set
	self.role = role or nil --Purpose of the traffic car; normal, police, other
	self.action = action or nil --Additional action for the AI; none, watch, chase
	self.lane = lane or 0 --Vehicle lane
	self.node = node or {} --Vehicle virtual node
	self.respawn = respawn or 1 --Respawn multiplier variable
	self.road = road or {} --Nearest road data
	self.timer = timer or {} --General timer for AI
	self.drive = drive or {} --Manual drive
	return self
end

local function ready()
	log("Z", "gull", "Traffic Tool ready")
end

local function getRoadScore(drivability, radius) --Gets the average value of road drivability and road radius
	return (drivability + (math.min(6 / 5, radius / 5))) / 2 --Rejects super wide roads
end

local function getLicenseText() --Creates a simple license plate text string
	local z = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z'}
	return z[math.random(1, #z)]..z[math.random(1, #z)]..z[math.random(1, #z)].."-"..math.random(0, 9)..math.random(0, 9)..math.random(0, 9)..math.random(0, 9)
end

local function pickSpawn(spawnPointMid, spawnGap) --Picks three possible spawn points and gives them scores; best one wins (future: try path based spawn point)
	local mainMap = map.getMap()
	local spawnPoint = {pos = nil, rot = nil, data = nil, p1 = nil, p2 = nil, dist = nil}
	local bestPoint = {}
	local angles = {quatFromDir(vec3(-0.577, 1, 0):normalized(), up), quatFromDir(vec3(0, 1, 0), up), quatFromDir(vec3(0.577, 1, 0):normalized(), up)} --30 degrees
	local roadScore = 0
	local score = 0
	local bestScore = 0
	local upScore = 0
	local a = 0
	if scatterCenter == true then --Randomizes or shifts the center point
		if mainMap.nodes[spawnPointMid.p1].links[spawnPointMid.p2].oneWay == true then
			a = 0.08 --Right side bias; makes highways a bit more fun
		else
			a = math.random(-132, 132) / 1000 --Random narrow angle
		end
		angles[2] = quatFromDir(vec3(a, 1, 0):normalized(), up)
	end
	for i, v in ipairs(angles) do
		spawnPoint.p1, spawnPoint.p2, spawnPoint.dist = map.findClosestRoad(vec3(spawnPointMid.pos) - vec3(spawnPointMid.rot):rotated(angles[i]) * spawnGap * -1)
		if i == 2 then --Center spawn point is favored and therefore given a small multiplier
			upScore = 1.15
		else
			upScore = 1
		end
		if spawnPoint.p1 then
			spawnPoint.data = mainMap.nodes[spawnPoint.p1].links[spawnPoint.p2]
			roadScore = getRoadScore(spawnPoint.data.drivability, mainMap.nodes[spawnPoint.p1].radius)
			score = math.ceil((100 - (spawnPoint.dist / 10)) * roadScore * upScore) --Formula for spawn point score
		end
		if score >= bestScore then
			bestScore = score
			for k, v in pairs(spawnPoint) do
				bestPoint[k] = v
			end
		end
	end
	return bestPoint
end

local function findSpawn(t) --Finds and returns a spawn point on the map
	local mainMap = map.getMap()
	local spawnPoint = {}
	local spawnPointMid = {}
	local gotSpawn = false
	local spawnGap = nil
	local loopCount = 0
	while gotSpawn == false do
		if loopCount == 0 then
			spawnPointMid.p1, spawnPointMid.p2, spawnPointMid.dist = map.findClosestRoad(vec3(player.pos) - vec3(player.vec) * (pointDist / 2) * -1) --Midway spawn point focus
		end
		if spawnPointMid.p1 then
			spawnPointMid.pos = mainMap.nodes[spawnPointMid.p1].pos
			spawnPointMid.rot = (vec3(spawnPointMid.pos) - vec3(player.pos)):normalized()
			spawnGap = ((pointDist * spawnGapFactor) / 2) + (player.vel * 2)  + (loopCount * 25) --Higher player velocity moves the spawn point further; loop moves it even further
			spawnPoint = pickSpawn(spawnPointMid, spawnGap)
		end
		if spawnPoint.p1 then
			gotSpawn = true --Temporarily set to true
			spawnPoint.pos = mainMap.nodes[spawnPoint.p1].pos
			spawnPoint.data = mainMap.nodes[spawnPoint.p1].links[spawnPoint.p2]
			for i, v in ipairs(traffic) do
				if traffic[i].obj.name ~= t.obj.name then
					if (spawnPoint.pos:toPoint3F() - traffic[i].pos):len() <= pointDist / 8 or spawnPoint.data.drivability < roadDrive then --Prevents spawn points inside other traffic cars or non roads
						gotSpawn = false
					end
				end
			end
			if (spawnPoint.pos:toPoint3F() - player.pos):len() <= pointDist * spawnGapFactor then --Prevents spawn points too close to player
				gotSpawn = false
			end
		end
		if loopCount >= 19 then --20 loops
			spawnPoint.p1, spawnPoint.p2, spawnPoint.dist = nil, nil, nil
			gotSpawn = true
			break
		end
		loopCount = loopCount + 1
	end
	return spawnPoint
end

local function findParallelRoad(spawnPoint, spawnGap) --Finds parallel roads to the left or right of the original road
	local mainMap = map.getMap()
	local tempPoint = {}
	local xPoint = {}
	local yPoint = {}
	local roadDir = 0
	if not spawnGap then
		spawnGap = 5
	end
	local spawnPosTemp = {mainMap.nodes[spawnPoint.p1].pos, mainMap.nodes[spawnPoint.p2].pos}
	local spawnGapTemp = {(player.pos - spawnPosTemp[1]:toPoint3F()):len(), (player.pos - spawnPosTemp[2]:toPoint3F()):len()}
	if spawnGapTemp[1] < spawnGapTemp[2] then --Figures out which spawn point from the pair is closest to the player so a proper rotation can be done
		tempPoint.pos = spawnPosTemp[1]
		tempPoint.rot = (vec3(spawnPosTemp[2]) - vec3(tempPoint.pos)):normalized()
	else
		tempPoint.pos = spawnPosTemp[2]
		tempPoint.rot = (vec3(tempPoint.pos) - vec3(spawnPosTemp[1])):normalized()
	end
	for i = 1, 2, 1 do
		if i == 1 then
			xPoint.pos = vec3(tempPoint.pos) - tempPoint.rot:z0():cross(up) * spawnGap * -1
		else
			xPoint.pos = vec3(tempPoint.pos) - tempPoint.rot:z0():cross(up) * spawnGap
		end
		xPoint.p1, xPoint.p2, xPoint.dist = map.findClosestRoad(xPoint.pos)
		xPoint.rot = (vec3(mainMap.nodes[xPoint.p1].pos) - vec3(mainMap.nodes[xPoint.p2].pos)):normalized()
		if xPoint.dist <= spawnGap / 1.3 and math.abs(tempPoint.rot:dot(xPoint.rot)) >= 0.8 then --Nearby side node found with similar direction
			roadDir = i
		else
			yPoint.pos = vec3(xPoint.pos) - tempPoint.rot * spawnGap * 1.5 * -1 --Slight distance boost
			yPoint.p1, yPoint.p2, yPoint.dist = map.findClosestRoad(yPoint.pos)
			yPoint.rot = (vec3(mainMap.nodes[yPoint.p1].pos) - vec3(mainMap.nodes[yPoint.p2].pos)):normalized()
			if yPoint.dist <= spawnGap / 1.3 and xPoint.p1 ~= yPoint.p1 and math.abs(xPoint.rot:dot(yPoint.rot)) >= 0.8 then --Nearby front node found with similar direction
				roadDir = i
			end
		end
	end
	return roadDir, quatFromDir(tempPoint.rot, up)
end

local function setSpawn(t) --Sets the new spawn point of a traffic car and rotates it logically
	local mainMap = map.getMap()
	local spawn = {pos = nil, rot = nil, posNew = nil, rotNew = nil, shiftOut = 0, shiftUp = 0.5}
	local spawnPoint = {pos = nil, rot = nil, data = nil, p1 = nil, p2 = nil, dist = nil}
	local spawnWay = 1
	local spawnHighway = 0
	local spawnRef = nil
	local laneShift = 1.8
	local colorSet = {}
	local colorValues = {}
	spawnPoint = findSpawn(t)
	if not spawnPoint.p1 then
		log("Z", "gull", "Warning: "..t.obj.name.." ("..t.obj.jbeam..") didn't get a new spawn point!")
	else
		spawn.pos = spawnPoint.pos
		spawn.roadScore = getRoadScore(spawnPoint.data.drivability, mainMap.nodes[spawnPoint.p1].radius)
		spawnHighway, spawn.rot = findParallelRoad(spawnPoint, 24)
		t.lane = 1
		if spawnPoint.data.oneWay == false then
			spawnWay = math.random(1, 2)
			if spawnHighway > 0 then
				spawnWay = spawnHighway
			else
				spawn.rot = quatFromDir(spawnPoint.pos - mainMap.nodes[spawnPoint.p2].pos, up) --Best way to get node rotation
			end
			if mainMap.nodes[spawnPoint.p1].radius <= 2.2 then --Road is really narrow (temporary solution)
				spawn.rot = quatFromDir(player.vec)
				spawnWay = 2
			end
		else
			spawnRef = spawnPoint.data.inNode
			if spawnRef == spawnPoint.p1 then
				spawn.rot = quatFromDir(spawnPoint.pos - mainMap.nodes[spawnPoint.p2].pos, up)
			else
				spawn.rot = quatFromDir(spawnPoint.pos - mainMap.nodes[spawnPoint.p1].pos, up)
			end
			if mainMap.nodes[spawnPoint.p1].radius >= 3 then --Road is wide enough to have at least two lanes
				t.lane = math.random(0, 1)
			end
		end
		if t.lane == 1 then
			spawn.shiftOut = laneShift * spawn.roadScore * (spawnWay * -2 + 3)
		else
			spawn.shiftOut = laneShift * spawn.roadScore * (t.lane * 2 - 1)
		end
		spawn.posNew = vec3(spawn.pos:toPoint3F()) - vec3(0, 1, 0):rotated(spawn.rot):z0():cross(up) * spawn.shiftOut
		if spawnWay == 1 then
			spawn.rotNew = quatFromDir(vec3(0, 1, 0):rotated(spawn.rot), up)
		else
			spawn.rotNew = quatFromDir(vec3(0, 1, 0):rotated(spawn.rot):rotated(quatFromDir(vec3(0, -1, 0), up)), up)
		end
		if math.abs(spawn.rotNew.x) >= 0.08 or math.abs(spawn.rotNew.y) >= 0.08 then --Hill spawn point workaround
			spawn.shiftUp = 1
		end
		t.obj:setPositionRotation(spawn.posNew.x, spawn.posNew.y, spawn.posNew.z + spawn.shiftUp, 0, 0, spawn.rotNew.z, spawn.rotNew.w + (math.random(1, 100) / 1000)) --The random addition sloppily fixes a vehicle rotation velocity bug
		log("Z", "gull", t.obj.name.." ("..t.obj.jbeam..") teleported to "..spawn.posNew.x..", "..spawn.posNew.y..", "..spawn.posNew.z)
		--core_vehicles.setPlateText(nil, t.obj:getId(), nil, nil, true)
		if t.damage.now >= damageUnit then
			t.obj:queueLuaCommand("vehicleResetted(false)")
		end
		t.obj:resetBrokenFlexMesh()
		if paintColorMode ~= "unchanged" and t.role ~= "police" and t.role ~= "ambulance" then
			if paintColorMode == "standard" then
				colorSet = paintColors[math.random(1, #paintColors)]
				colorValues = {colorSet[1], colorSet[2], colorSet[3], colorSet[4]}
			elseif paintColorMode == "random" then
				colorValues = {math.random(), math.random(), math.random(), math.random() + 1}
			end
			t.obj.color = ColorF(colorValues[1], colorValues[2], colorValues[3], colorValues[4]):asLinear4F()
			if t.obj.jbeam == "fullsize" or t.obj.jbeam == "roamer" then --Common vehicles with dual paint styles
				t.obj.colorPalette0 = t.obj.color
			end
		end
		t.road.drive = spawnPoint.data.drivability
		t.road.radius = mainMap.nodes[spawnPoint.p1].radius
		t.road.score = spawn.roadScore
		t.road.parallel = spawnHighway
	end
	t.state = "set"
	t.delay = false
end

local function getNumberList(bunch, total) --Returns a list of non-repeating random integers
	local n = {}
	local l = {}
	local key = 0
	for i = 1, total, 1 do
		table.insert(n, i)
	end
	for i = 1, bunch, 1 do
		key = math.random(1, #n)
		table.insert(l, n[key])
		table.remove(n, key)
	end
	return l
end

local function setQuantity(quantity)
	if not quantity then
		log("Z", "gull", "Quantity was nil!")
	else
		trafficTotal = quantity
	end
end

local function setGroup(int)
	local groupList = {"normal", "allBase", "allAny", "light", "heavy", "america", "europe", "asia", "americaPolice", "europePolice", "allPolice", "automation", "copyPlayer", "custom"}
	groupName = groupList[int]
	for k, v in pairs(tGroup) do
		if groupName == k then
			group = v
			ui_message("Spawn group set: "..groupName, 6)
			log("Z", "gull", "Group: "..groupName)
		end
	end
	if int > tableSize(groupList) then
		log("Z", "gull", "Group ID doesn't exist!")
	end
end

local function setCustom(list) --Sets a custom traffic group for the spawner
	local tList = {}
	if type(list) == "string" then
		tList = {list}
	elseif type(list) == "table" then
		tList = list
	else
		tList = {"pigeon/600.pc"}
	end
	tGroup.custom = tList
	group = tGroup.custom
	ui_message("Custom data ready", 6)
	log("Z", "gull", "Custom data ready")
end

local function setAsPolice(config) --Sets the current player vehicle to act as police
	if not config then
		config = "nil"
		local tempPlayer = be:getPlayerVehicle(0)
		if tempPlayer.partConfig then
			config = string.gsub(tempPlayer.partConfig, "vehicles/", "")
		end
	end
	table.insert(tGroup.allPolice, config)
	ui_message("Police car set: "..config, 6)
	log("Z", "gull", "Police vehicle: "..config)
end

local function getAllVehicles(class) --Gets all vehicles depending on group type
	if not class then
		class = "allAny"
	end
    local vList = {}
	local name = nil
	local json = nil
    for i, v in ipairs(FS:findFiles("/vehicles", "*", 0, false, true)) do
        if v ~= "/vehicles/common" then
			json = jsonReadFile(v.."/info.json")
			if json then
				if json["Type"] == "Car" or json["Type"] == "Truck" or json["Type"] == "Automation" then
					if class == "allAny" then
						for o, w in ipairs(FS:findFiles(v, "*.pc", 0, true, false)) do
							table.insert(vList, w)
						end
					else	
						if string.byte(v) == 47 then
							name = string.gsub(v, "/vehicles/", "")
						else
							name = string.gsub(v, "vehicles/", "")
						end
						name = string.match(name, "%w*_*%w*_*%w*_*%w*_*%w*") --This seems wrong
						if class == "allBase" then
							table.insert(vList, name)
						end
						if class == "automation" and json["Type"] == "Automation" then
							table.insert(vList, name)
						end
					end
				end
			end
        end
    end
	return vList
end

local function checkVehicle(vehicle) --Gets the vehicle type from the .json file
	local vType = nil
	local jbName = scenetree.findObject(vehicle).jbeam
	local json = jsonReadFile("/vehicles/"..jbName.."/info.json")
	if json then
		if json["Type"] then
			vType = string.lower(json["Type"])
		end
	else
		vType = "unknown" --Hmm
	end
	return vType
end

local function getTrafficTable() --Returns a table with valid traffic vehicles
	local otherVehicles = scenetree.findClassObjects("BeamNGVehicle")
	local validTypes = {"car", "truck", "automation"}
	local delete = {}
	for i, v in ipairs(otherVehicles) do
		local vType = checkVehicle(v)
		local isValid = false
		for o, w in ipairs(validTypes) do
			if vType == w then
				isValid = true
			end
		end
		if otherVehicles[i] == be:getPlayerVehicle(0).name or isValid == false then
			table.insert(delete, i)
		end
	end
	for i = #delete, 1, -1 do
		table.remove(otherVehicles, delete[i])
	end
	return otherVehicles
end

local function findClosestTraffic() --Finds the nearest traffic vehicle
	local tempDist = farDist
	local t = nil
	for i, v in ipairs(traffic) do
		if traffic[i].dist.now <= tempDist then
			tempDist = traffic[i].dist.now
			t = traffic[i]
		end
	end
	return t
end

local function clearVehicles() --Deletes traffic vehicles
	local otherVehicles = getTrafficTable()
	for i, v in ipairs(otherVehicles) do
		scenetree.findObject(otherVehicles[i]):delete()
	end
	trafficReady = false
	if #otherVehicles > 0 then
		ui_message("All AI vehicles cleared", 6)
	end
end

local function setInitSpawn(active, total) --Sets the spawn points of the initial vehicles
	local list = {}
	if total <= #group then
		list = getNumberList(total, #group)
	else
		for i = 1, total, 1 do
			table.insert(list, math.random(1, #group))
		end
	end
	local groupData = {}
	local inData = {}
	local player0 = be:getPlayerVehicle(0)
	local mainColor = nil
	local spawnGap = 12
	if groupName == "heavy" then
		spawnGap = 16
	end
	if groupName == "copyPlayer" then
		tGroup.copyPlayer = {player0.partConfig}
		mainColor = tostring(player0.color)
		mainColor = stringToTable(mainColor, ", ")
		mainColor = ColorF(mainColor[1], mainColor[2], mainColor[3], mainColor[4]) --Phew!
		for i = 1, total, 1 do
			table.insert(groupData, {player0.jbeam, player0.partConfig, mainColor})
		end
	else
		for i = 1, total, 1 do
			inData = {}
			local name = group[list[i]]
			local jbName = nil
			local nameSpec = nil
			local json = nil
			mainColor = "1 1 1 1.5"
			if string.byte(name) == 47 then --Temporarily remove "/vehicles/" from string
				name = string.gsub(name, "/vehicles/", "")
			else
				name = string.gsub(name, "vehicles/", "")
			end
			if string.find(name, ".pc") then
				nameSpec = "pc"
			end
			if nameSpec == "pc" then
				jbName = string.match(name, "%w*_*%w*_*%w*_*%w*_*%w*")
				inData[1] = jbName
				inData[2] = "vehicles/"..name
			else
				jbName = name
				inData[1] = jbName
				inData[2] = nil
			end
			json = jsonReadFile("/vehicles/"..jbName.."/info.json")
			if json then
				local colorTable = json["colors"]
				local colorName = json["default_color"]
				local fileName = nil
				if nameSpec == "pc" then
					name = string.gsub(name, jbName.."/", "")
					name = string.gsub(name, ".pc", "")
					fileName = "/vehicles/"..jbName.."/info_"..name..".json"
					json = jsonReadFile(fileName)
					if json then
						colorName = json["default_color"]
					end
				end
				for k, v in pairs(colorTable) do
					if colorName == k then
						mainColor = v
					end
				end
			end
			mainColor = stringToTable(mainColor, " ")
			inData[3] = ColorF(mainColor[1], mainColor[2], mainColor[3], mainColor[4])
			table.insert(groupData, inData)
		end
	end
	for i = 1, total, 1 do
		local spawnPos = vec3(player0:getPosition()) - vec3(player0:getDirectionVector()) * spawnGap * (active + i) * -1
		local spawnRot = vec3(player0:getDirectionVector())
		spawn.spawnVehicle(groupData[i][1], groupData[i][2], vec3(spawnPos):toPoint3F(), quatFromDir(spawnRot, up), groupData[i][3])
	end
end

local function spawnVehicles() --Spawns the required amount of vehicles
	if trafficTotal >= 1 then
		local player0 = be:getPlayerVehicle(0)
		local otherVehicles = getTrafficTable()
		if groupName == "allBase" or groupName == "allAny" or groupName == "automation" then
			tGroup[groupName] = getAllVehicles(groupName)
			group = tGroup[groupName]
		end
		local trafficNeeded = trafficTotal - #otherVehicles
		if trafficNeeded > 0 then
			setInitSpawn(#otherVehicles, trafficNeeded)
			be:enterVehicle(0, player0)
			if trafficNeeded >= 12 then
				ui_message("Spawned "..trafficNeeded.." new AI vehicles; that's quite insane!", 12)
			else
				ui_message("Spawned "..trafficNeeded.." new AI vehicles", 6)
			end
		else
			ui_message("There are already "..trafficTotal.." AI vehicles on the map!", 6)
		end
	else
		log("Z", "gull", "Traffic quantity set to 0!")
	end
end

local function activateVehicles() --Processes and activates vehicles on the map, and enables the main script to run
	local playerName = nil
	local playerObj = nil
	local otherVehicles = getTrafficTable()
	if #player == 0 then
		playerObj = be:getPlayerVehicle(0)
		playerName = playerObj.name
	else
		playerObj = scenetree.findObject(playerName)
	end
	player = Player:new(playerObj, nil, 0, nil, 0, {name = nil, pair = false}, 0, 0, nil, {police = 0, evade = 0}) --obj, pos, vel, vec, dist, pair, weight, damage, state, timer
	traffic = {}
	if #otherVehicles < 1 then
		trafficReady = false
	else
		for i, v in ipairs(otherVehicles) do
			traffic[i] = Traffic:new(scenetree.findObject(otherVehicles[i]), nil, 0, nil, speedLimit, {}, {}, "set", "normal", "none", 1, {}, spawnPassFactor, {}, {}, {}) --obj, pos, vel, vec, speed, dist, damage, state, role, action, lane, node, respawn, road, timer, drive
			traffic[i].road = {drive = 1, radius = 5, score = 1, parallel = 0}
			traffic[i].timer = {wait = 0, active = 0, path = 0, delay = false}
			traffic[i].drive = {mode = "none", timer = 0, data = {}}
			traffic[i].obj:queueLuaCommand("extensions.load('aitt')") --Custom AI extension
		end
		trafficReady = true
		policeReady = false
		for i, v in ipairs(traffic) do
			traffic[i].obj:queueLuaCommand("aitt.setSpeedMode('limit')")
			traffic[i].obj:queueLuaCommand("aitt.resetLearning(true)")
			for o, w in ipairs(tGroup.allPolice) do
				if traffic[i].obj.partConfig == "vehicles/"..tGroup.allPolice[o] then
					traffic[i].role = "police"
					policeReady = true
				end
			end
			if string.find(traffic[i].obj.partConfig, "police") then
				traffic[i].role = "police"
				policeReady = true
			end
			if string.find(traffic[i].obj.partConfig, "ambulance") then
				traffic[i].role = "ambulance"
			end
		end
		ui_message("Traffic activated!", 6)
	end
end

local function validateVehicles() --Validates all vehicles; prevents game crashes and stuff
	local allVehicles = scenetree.findClassObjects("BeamNGVehicle")
	if #allVehicles - 1 < #traffic and #traffic > 0 then
		activateVehicles()
	else
		if #allVehicles - 1 == 0 then
			trafficReady = false
		end
	end
end

local function togglePolice()
	policeReady = not policeReady
	log("Z", "gull", "Police mode: "..tostring(policeReady))
end

local function toggleHeadlights()
	headlightsMode = not headlightsMode
	log("Z", "gull", "Headlights mode: "..tostring(headlightsMode))
end

local function toggleDamageKill()
	damageMode = not damageMode
	log("Z", "gull", "Damage mode: "..tostring(damageMode))
end

local function setSpeedLimit(speedLimitValue)
	if not speedLimitValue then
		log("Z", "gull", "Speed limit value was nil!")
	else
		speedLimit = speedLimitValue / speedConvert[speedUnits] --To m/s
	end
end

local function toggleUnits(metricFlag)
	if not metricFlag then
		log("Z", "gull", "Unit value was nil!")
	else
		speedUnits = metricFlag
		setSpeedLimit(speedLimit * speedConvert[3 - speedUnits]) --Sloppy
	end
end

local function setRespawnRate(respawnRate)
	if not respawnRate then
		log("Z", "gull", "Respawn rate value was nil!")
	else
		spawnFactor = respawnRate
		spawnGapFactor = -respawnRate / 5 + 1.2 --Variable respawn distance multiplier
	end
end

local function changeColorMode(colorMode)
	if not colorMode then
		log("Z", "gull", "Color mode value was nil!")
	else
		if colorMode == 1 then
			paintColorMode = "unchanged"
		elseif colorMode == 2 then
			paintColorMode = "standard"
		elseif colorMode == 3 then
			paintColorMode = "random"
		end
	end
end

local function changeBotMode(botMode)
	if not botMode then
		log("Z", "gull", "AI mode value was nil!")
	else
		if botMode == 1 then
			trafficMode = "random"
			laneMode = "on"
		elseif botMode == 2 then
			trafficMode = "flee"
			laneMode = "off"
		elseif botMode == 3 then
			trafficMode = "chase"
			laneMode = "off"
		elseif botMode == 4 then
			trafficMode = "disabled"
			laneMode = "off"
		end
		if trafficReady == true then
			for i, v in ipairs(traffic) do
				traffic[i].obj:queueLuaCommand("aitt.setMode('"..trafficMode.."')")
				traffic[i].obj:queueLuaCommand("aitt.driveInLane('"..laneMode.."')")
				if trafficMode == "disabled" then
					traffic[i].obj:queueLuaCommand("extensions.unload('aitt')")
					trafficReady = false
				end
			end
		end
	end
end

local function toggleSmartSpeed(smartMode)
	if not smartMode then
		log("Z", "gull", "Smart AI mode value was nil!")
	else
		if smartMode == 1 then
			smartSpeed = true
		else
			smartSpeed = false
		end
	end
end

local function toggleAltRespawn(altRespawn)
	if not altRespawn then
		log("Z", "gull", "Alt respawn value was nil!")
	else
		if altRespawn == 1 then
			spawnPassFactor = 1
		else
			spawnPassFactor = 1.6
		end
	end
end

local function getSmartSpeed(t) --Sets a situational speed or action for the AI (needs more work)
	local speed = 0
	local mode = "normal"
	local target = nil
	local policeDist = farDist
	local randomAction = 0
	local step = 10
	local speedWatch = 10
	local speedFactor = 0
	local tempDist = 0
	local pointDist = (2.5 + math.min(1.5, t.vel / 5)) * t.road.score --Point area changes
	if player.pair.lock == true and t.lane == 1 then --Experimental thing here
		pointDist = 2 * t.road.score
		t.lane = 0.8
		t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
	end
	local frontDist = 5
	local frontValue = math.max(1, t.vel)
	local frontPos = nil
	local tempPos = vec3(t.pos)
	for i = 2, 0, -1 do
		frontPos = (tempPos - t.vec * (frontDist + frontValue * i) * -1):toPoint3F() --AI front position points are done in "beads"
		for o, w in ipairs(traffic) do
			if traffic[o].obj.name ~= t.obj.name then
				tempDist = (frontPos - traffic[o].pos):len()
				if tempDist <= pointDist then
					step = i
					target = "traffic"
					speedWatch = traffic[o].vel
				end
			end
		end
		tempDist = (frontPos - player.pos):len()
		if tempDist <= pointDist and i < step then
			step = i
			target = "player"
			speedWatch = player.vel
		end
	end
	if step <= 2 then
		if target == "player" and player.vec:dot(t.vec) <= -0.9 then --Player facing AI head to head; AI will brake a bit harder
			speedWatch = 0
		end
		speedFactor = 0.75 + (step / 15) + (math.max(0.1, speedWatch) / math.max(0.1, t.vel)) / 6 --Sloppy formula here
		speed = math.min(speedLimit, (speedWatch * speedFactor)) --Formula for distance based speed
	else
		if t.timer.wait > 0 then
			if t.speed < 1 then
				speed = t.speed + math.min(0.15, speedWatch / 10) --Somewhat gradual acceleration
			else
				speed = t.speed * 1.08
			end
		else
			speed = speedLimit * t.road.score
		end
	end
	for i, v in ipairs(traffic) do
		if traffic[i].obj.name ~= t.obj.name then
			if t.road.radius <= 2.5 and t.road.parallel == 0 then
				if ((t.node[2]:toPoint3F() - traffic[i].node[2]:toPoint3F()):len() <= 5 or (t.node[2]:toPoint3F() - player.pos):len() <= 10) then
					mode = "pullOver"
				end
			end
			if t.role ~= "police" and traffic[i].role == "police" and traffic[i].action ~= "none" then
				if (t.node[3]:toPoint3F() - traffic[i].pos):len() <= policeDist then
					policeDist = (t.node[3]:toPoint3F() - traffic[i].pos):len()
				end
			end
		end
	end
	if zSiren > 0 then --If player siren is on
		local frontGap = readyDist / 2
		local tempObj = nil
		frontPos = vec3(player.pos) - player.vec * frontGap * -1
		tempDist = (t.pos - frontPos:toPoint3F()):len()
		if player.vel > crawlSpeed then
			player.pair.name = t.obj.name
			player.pair.lock = false
		else
			if player.pair.lock == false then
				tempObj = findClosestTraffic()
				player.pair.name = tempObj.obj.name
				player.pair.lock = true
			end
		end
		if t.role ~= "police" and player.pair.name == t.obj.name then
			if tempDist <= frontGap then
				policeDist = tempDist
				if t.timer.path == 0 and math.random(1, 10) == 1 then
					t.role = "suspect"
				end
			end
		end
	end
	if policeDist <= 50 then
		mode = "pullOverPolice"
	else
		if t.timer.path ~= 0 then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			t.timer.path = 0
		end
	end
	if mode == "bypass" then --Modes use the path timer to process actions (better method needed)
		if t.timer.path == 0 then
			t.timer.path = 0.1
			t.lane = 0.5
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
		elseif t.timer.path >= 5 then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			t.timer.path = 0
		end
	elseif mode == "pullOver" then
		if t.timer.path == 0 then
			t.timer.path = 0.1
			t.lane = 1.5
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
		elseif t.timer.path >= 5 then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			t.timer.path = 0
		end
		speed = math.max(crawlSpeed, speed - t.timer.path * 1.5)
	elseif mode == "pullOverPolice" then
		if t.timer.path == 0 then
			t.timer.path = 0.1
			t.lane = 1.5
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
		elseif t.timer.path >= 8 then
			if t.role == "suspect" then --Traffic vehicle runs from police!
				t.action = "flee"
				t.obj:queueLuaCommand("aitt.setMode('flee')")
				t.obj:queueLuaCommand("aitt.driveInLane('off')")
				t.respawn = 0.1
				t.timer.path = 0
			end
		elseif t.timer.path >= 60 then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			t.timer.path = 0
		end
		speed = math.max(0.01, speed - t.timer.path * 3)
	end
	if t.timer.path ~= 0 then
		t.timer.path = t.timer.path + timeTick
	end
	return speed
end

local function driveScript(t) --Custom AI inputs based on certain situations (unused)
	if t.drive.mode ~= "reset" then
		t.obj:queueLuaCommand("input.event('steering', "..t.drive.data.s..", 1)")
		t.obj:queueLuaCommand("input.event('throttle', "..t.drive.data.t..", 2)")
		t.obj:queueLuaCommand("input.event('brake', "..t.drive.data.b..", 2)")
		t.obj:queueLuaCommand("input.event('parkingbrake', "..t.drive.data.pb..", 2)")
	else
		t.drive.mode = "none"
		t.drive.timer = 0
		t.drive.data = {s = 0, t = 0, b = 0, pb = 0}
		t.obj:queueLuaCommand("aitt.setMode('"..trafficMode.."')")
	end
end

local function updateRoad(t) --Updates the road data of the vehicle
	local mainMap = map.getMap()
	local point = {}
	point.p1, point.p2, point.dist = map.findClosestRoad(vec3(t.node[1]))
	if point.p1 then
		if t.lane == 0 and mainMap.nodes[point.p1].links[point.p2].oneWay == false then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
		end
		t.road.drive = mainMap.nodes[point.p1].links[point.p2].drivability
		t.road.radius = mainMap.nodes[point.p1].radius
		t.road.score = getRoadScore(t.road.drive, t.road.radius)
		t.road.parallel = findParallelRoad(point, 12)
	end
end

local function chasePlayer(t) --New method for AI to chase player
	local frontGap = readyDist / 2
	local frontPos = vec3(t.pos) - t.vec * frontGap * -1 --Point ahead of vehicle; chase mode trickery
	local frontDist = (player.pos - frontPos:toPoint3F()):len()
	if t.action == "watch" then
		t.speed = math.max(crawlSpeed, player.vel * ((readyDist / math.min(readyDist, t.dist.now)) / 15 + 0.4)) --Formula for "chase follow" speed
		if frontDist <= frontGap + math.max(5, player.vel / 4) or player.vel <= crawlSpeed then
			if t.role == "police" then
				t.respawn = math.max(0, policeSpawn)
			else
				t.respawn = 1
			end
			t.speed = 100
			t.action = "chase"
			t.obj:queueLuaCommand("aitt.driveInLane('off')")
			t.obj:queueLuaCommand("aitt.setMode('chase')")
		end
	end
end

local function simulatePolice() --Various logic for while the police car is chasing the player
	local chaseDist = readyDist / 2
	local arrestDist = farDist
	for i, v in ipairs(traffic) do
		if traffic[i].role == "police" and traffic[i].action ~= "none" then
			if traffic[i].dist.now <= arrestDist then
				arrestDist = traffic[i].dist.now
			end
		end
	end
	if player.timer.police >= arrestTime then
		arrestDist = 0
	end
	if arrestDist <= chaseDist then 
		if player.vel <= crawlSpeed then
			if math.floor(player.timer.police * 10) / 10 == arrestTime then
				player.obj:queueLuaCommand("ai.setMode('stopping')")
				for i, v in ipairs(traffic) do
					if traffic[i].role == "police" then
						traffic[i].obj:queueLuaCommand("aitt.setMode('stopping')")
						traffic[i].obj:queueLuaCommand("electrics.set_lightbar_signal(1)")
					end
				end
				ui_message("You got busted!", 5)
			elseif math.floor(player.timer.police * 10) / 10 == arrestTime + 5 then
				player.obj:queueLuaCommand("ai.setMode('disabled')") --Warning, bug with gearbox mode
				for i, v in ipairs(traffic) do
					if traffic[i].role == "police" then
						traffic[i].obj:queueLuaCommand("aitt.setMode('random')")
						traffic[i].obj:queueLuaCommand("electrics.set_lightbar_signal(0)")
						traffic[i].obj:queueLuaCommand("aitt.driveInLane('on')")
						traffic[i].speed = speedLimit
						traffic[i].action = "none"
					end
				end
				ui_message("You can drive away now...", 5)
				policeSpawn = 1.2
				policeAlert = false
			end	
			player.timer.police = player.timer.police + timeTick
		else
			player.timer.police = 0
		end
		player.timer.evade = 0
	else
		if player.timer.evade >= evadeTime then
			for i, v in ipairs(traffic) do
				if traffic[i].role == "police" then
					traffic[i].obj:queueLuaCommand("aitt.setMode('random')")
					traffic[i].obj:queueLuaCommand("electrics.set_lightbar_signal(0)")
					traffic[i].obj:queueLuaCommand("aitt.driveInLane('on')")
					traffic[i].speed = speedLimit
					traffic[i].action = "none"
				end
			end
			ui_message("You successfully evaded the police!", 5)
			policeSpawn = 1.2
			policeAlert = false
		end
	end
	player.timer.evade = player.timer.evade + timeTick
end

local function simulateTraffic() --Various logic for traffic; also handles when to respawn traffic
	player.pos = player.obj:getPosition()
	player.vel = map.objects[player.obj:getID()].vel:length()
	player.vec = vec3(player.obj:getDirectionVector())
	player.damage = map.objects[player.obj:getID()].damage
	player.obj:queueLuaCommand("obj:queueGameEngineLua('zSiren = ('..electrics.values.lightbar..')')") --Player siren state
	local timeDay = core_environment.getTimeOfDay()
	for i, v in ipairs(traffic) do
		local t = traffic[i]
		t.pos = t.obj:getPosition()
		t.dist.now = (player.pos - t.pos):len()
		t.vel = map.objects[t.obj:getID()].vel:length()
		t.vec = vec3(t.obj:getDirectionVector())
		t.damage.now = map.objects[t.obj:getID()].damage
		t.node[1] = vec3(t.pos) - vec3(t.vec) * 5 * -1 --Temporary idea
		t.node[2] = vec3(t.pos) - vec3(t.vec) * 20 * -1
		t.node[3] = vec3(t.pos) - vec3(t.vec) * 50 * -1
		if t.state == "set" then
			if t.timer.delay == true then --Slight one tick delay for AI mode change (prevents FPS drop spike? Test this...)
				if trafficMode == "chase" then
					if t.role == "police" then
						policeAlert = true
					end
					t.obj:queueLuaCommand("aitt.setMode('random')")
					t.obj:queueLuaCommand("aitt.driveInLane('on')")
					t.action = "watch"
				else
					t.obj:queueLuaCommand("aitt.setMode('"..trafficMode.."')")
					t.obj:queueLuaCommand("aitt.driveInLane('"..laneMode.."')")
				end
				if trafficMode == "random" then
					t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
					t.obj:queueLuaCommand("aitt.setAggression(0.75)")
				else
					t.obj:queueLuaCommand("aitt.setAggression(1.1)")
				end
				if t.role == "suspect" then
					t.role = "normal"
					t.action = "none"
				end
				if t.role == "police" and t.action ~= "none" then
					policeSpawn = policeSpawn - 0.025 --Police respawn multiplier decreases every time a new unit appears
					t.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
					t.obj:queueLuaCommand("aitt.setAggression(1.2)")
					t.speed = speedLimit
					t.lane = 0.8
					t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
					t.action = "watch"
				end
				if timeDay.time then
					if headlightsMode == true and timeDay.time >= 0.22 and timeDay.time <= 0.77 then
						t.obj:queueLuaCommand("electrics.setLightsState(2)")
					else
						t.obj:queueLuaCommand("electrics.setLightsState(0)")
					end
				end
				t.obj:queueLuaCommand("electrics.set_warn_signal(0)")
				t.state = "normal"
				t.damage.bad = false
				t.respawn = 1
				t.timer = {wait = 0, active = 0, path = 0, delay = false}
			end
			t.dist.prev = t.dist.now
			t.timer.delay = true
		end
		if t.timer.active < 1 then --Had to really force these variables
			t.damage.init = t.damage.now
			t.damage.prev = t.damage.now
		end
		if t.timer.active >= 1 and t.dist.now <= nearDist then
			if t.action == "none" then
				t.respawn = spawnPassFactor
			end
		end
		if t.dist.now <= readyDist / 2 then
			if t.role == "police" and t.action == "none" and policeReady == true then
				if player.vel >= speedLimit * 1.25 or (t.dist.now <= nearDist / 2 and t.damage.now - t.damage.prev >= damageUnit / 2) then --Minor offense
					t.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
					t.action = "watch"
					policeAlert = true
				end
				if player.vel >= speedLimit * 1.4 or (t.dist.now <= nearDist / 2 and t.damage.now - t.damage.prev >= damageUnit * 2) then --Major offense
					for o, w in ipairs(traffic) do
						if traffic[o].role == "police" and traffic[o].action == "none" then
							traffic[o].obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
							traffic[o].action = "watch"
							policeAlert = true
						end
					end
				end
			end
			t.damage.prev = t.damage.now
		end
		if spawnFactor > 0 and t.respawn > 0 and t.timer.active >= 1 and t.dist.now >= math.max(readyDist / 2, (farDist - ((t.dist.now - t.dist.prev) * (1 - math.max(0, (farDist / 2 - t.dist.now * 1.1) / farDist)) * distMultiplier * spawnFactor * t.respawn))) then --New formula for smart traffic respawning
			t.state = "ready"
		end
		if t.state == "ready" then
			if t.delay == false then
				t.obj:queueLuaCommand("aitt.setMode('disabled')")
			elseif t.delay == true then
				if queue > 0 then
					setSpawn(t)
					queue = -2 --Artificial delay for next spawn test; prevents many multiple teleports and lag (is this needed anymore?)
				end
			end
			t.delay = true
		end
		t.dist.prev = t.dist.now
		if t.action ~= "chase" and t.action ~= "flee" then
			if t.state ~= "set" and trafficMode == "random" then
				if t.drive.timer == 0 then
					if smartSpeed == true then
						t.speed = getSmartSpeed(t)
					else
						t.speed = speedLimit * t.road.score
					end
				end
				if damageMode == true then
					if t.damage.now - t.damage.init >= damageUnit * 2 then --Traffic cars try to stop driving if heavily damaged
						t.speed = (damageUnit * 10) / (t.damage.now - t.damage.init) --damage.init works around a false damage value bug for some vehicles
					end
					if t.damage.now - t.damage.init >= damageUnit * 4 then
						if t.damage.bad == false and math.random(1, 10) == 10 then
							t.respawn = 0.8 --Delays respawn in case the player wants to look back at the site of the wreck
							t.lane = 1.5
							t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
							t.obj:queueLuaCommand("electrics.set_warn_signal(1)")
							t.damage.bad = true
						end
					end
				end
			else
				t.speed = 100
			end
			if t.speed <= crawlSpeed * 2 then --Waiting drivers
				t.timer.wait = t.timer.wait + timeTick
			else
				t.timer.wait = 0
			end
		else
			t.speed = 100
		end
		if t.action ~= "none" then
			chasePlayer(t)
		end
		if (math.floor(t.timer.active * 10) / 10) % 1 == 0 then --Update road data every second
			updateRoad(t)
		end
		t.obj:queueLuaCommand("aitt.setSpeed("..tostring(t.speed)..")") --Main speed setter
		t.timer.active = t.timer.active + timeTick
	end
	if policeReady == true and policeAlert == true then
		simulatePolice()
	end
	queue = queue + 1 --This gets updated every tick to prevent two cars from being teleported at the same time
	if queue > #traffic then
		queue = 1
	end
end

local function onPreRender(dt, dtSim)
	loopTimer = loopTimer + dt
	if trafficReady == true then
		for i, v in ipairs(traffic) do
			traffic[i].obj:queueLuaCommand("aitt.updateGFX("..dtSim..")")
			if traffic[i].drive.mode ~= "none" then
				driveScript(traffic[i])
			end
		end
	end
	if loopTimer >= timeTick then --Runs on ticks instead of real time; similar to onRaceTick in scenarios (future: remove this method?)
		validateVehicles()
		if trafficReady == true then
			simulateTraffic()
		end
		loopTimer = 0
	end
end

--//Look into app parameters reset after using Graphics menu
--//Improve initial spawn method?
--//Hardcode West Coast highway spawning? Or wait until BeamNG update?
--//Get current game mode?

M.onPreRender = onPreRender

M.ready = ready
M.setQuantity = setQuantity
M.setGroup = setGroup
M.setCustom = setCustom
M.setAsPolice = setAsPolice
M.spawnVehicles = spawnVehicles
M.clearVehicles = clearVehicles
M.activateVehicles = activateVehicles
M.toggleUnits = toggleUnits
M.togglePolice = togglePolice
M.toggleHeadlights = toggleHeadlights
M.toggleDamageKill = toggleDamageKill
M.setSpeedLimit = setSpeedLimit
M.setRespawnRate = setRespawnRate
M.changeColorMode = changeColorMode
M.changeBotMode = changeBotMode
M.toggleSmartSpeed = toggleSmartSpeed
M.toggleAltRespawn = toggleAltRespawn

return M